/*  FILENAME: boxview.c

    AUTHOR:  Kelvin Thompson, 10/87

    DESCRIPTION:  Central routines for boxview program.
*/

#define CENTRAL
#include <stdio.h>
#include <math.h>
#include <gl.h>
#include <device.h>
#include <get.h>
#include "ktypes.h"
#include "idraw.h"

/* programs in this file */
extern main();
extern void init_iris();
extern void init_globs();
extern void st_draw();
extern void readystate();

/* programs in 'rotate.c' */
extern void st_axis();

/* programs in 'wm.c' */
extern WINDOW *open_wdw();

/* programs in 'view.c' */
extern VIEW *open_view();

/* local globals */
UNS16 objplanes;




/*  *********************
    *                   *
    *     main          *
    *                   *
    *********************

    DESCRIPTION:
*/

main ( argc, argv )
int argc;
char **argv;
{
/* initialize */
init_iris();
init_globs();
init_manip();
init_view();
init_rot();
init_obj( argc, argv );

/* open main window */
w0ptr = open_wdw( wminx, wmaxx, wminy, wmaxy );
v0ptr = open_view( w0ptr );
open_xh( v0ptr );

/* initialize states */
init_states();
readystate();

while ( TRUE )
  {
  if ( !do_state() ) break;
  writemask( 0xffffffff );
  swapbuffers();
  }

/* shut down */
close_wdw( w0ptr );
close_states();
gflush();
finish();
ginit();
gexit();
}



/*  *********************
    *                   *
    *     init_iris     *
    *                   *
    *********************

    DESCRIPTION: Initialize iris and and a couple of globals.

    GLOBALS:
      mexon -- boolean indicating that we're running under 'mex'
*/

/* color table */
#define LAST_ENTRY  0xfff
typedef struct CLR_ENTRY
  {
  Colorindex  idx;
  short    r,g,b;
  } CLR_ENTRY;
static CLR_ENTRY s0_entry[] = 
  {
  { s0_2CURSOR,    255,   0,   0 },
  { s0_3WDW,       200, 200, 200 },
  { s0_2WDW,       255, 255, 255 },
  { s0_AXES,       100, 100, 255 },
  { s0_AXNAMES,    200,   0,   0 },
  { s0_XH_EDGE,    180, 180,   0 },
  { LAST_ENTRY,      0,   0,   0 }
  };
static CLR_ENTRY s1_entry[] = 
  {
  { s1_BLACK,        0,   0,   0 },
  { s1_BACKG,        0,   0,   0 },
  { s1_VW_BACKG,     0,   0,   0 },
  { s1_VW_LINES,     0,   0, 255 },
  { s1_VW_GRABS,   250,   0,   0 },
  { s1_VW_BUTT,    200,   0,   0 },
  { s1_VW_AREA,    100, 100, 100 },
  { s1_SPLAT,      200, 200,   0 },
  { LAST_ENTRY,      0,   0,   0 }
  };

void init_iris ( )
{
register i;
register struct CLR_ENTRY *clr;
int buffbits;
unsigned maxcolor;

/* mex preferences */
foreground();
prefposition( 0, XMAXSCREEN, 0, YMAXSCREEN );

/* start & init graphics */
#ifdef FOURSIGHT
  winopen( "boxview" );
#else
  getport( "boxview" ); 
#endif
greset();

/* see if we're running mex */
getsize( &wmaxx, &wmaxy );
#ifdef FOURSIGHT
  mexon = FALSE;
#else
  mexon = ( wmaxx != 1024  ||  wmaxy != 1024 );
#endif

/* basic initialization */
onemap();
doublebuffer();
swapinterval(1);
gconfig();
setvaluator( MOUSEX, (XMAXSCREEN+1)/2, 0, XMAXSCREEN );
setvaluator( MOUSEY, (YMAXSCREEN+1)/2, 0, YMAXSCREEN );

/* clear everything */
color( BLACK );
clear();
swapbuffers();
color( BLACK );
clear();

/* enqueue all keyboard buttons */
for ( i=BUT0; i<MAXKBDBUT; i++ )
  qdevice( i );

/* load the grab font */
defrasterfont( GRAB_FONT_IDX, GRAB_MAX_HEIGHT, 
  NUM_GRAB_CHARS+1, grabchar, NUM_GRAB_RASTS, grabrast );

/* allocate the bitplane segments */
buffbits = cm_reset();
objplanes = buffbits - 3;
seg1 = cm_alloc( objplanes ) & 0xffff;
maxcolor = cm_alloc( 3 );
seg0 = maxcolor & 0xffff;
maxcolor = (1<<objplanes) - 1;

/* color segment 1 */
for ( clr = s1_entry;  clr->idx != LAST_ENTRY;  clr++ )
  cm_mapcolor( seg1, clr->idx, clr->r, clr->g, clr->b );

/* color segment 0 */
for ( clr = s0_entry;  clr->idx != LAST_ENTRY;  clr++ )
  cm_mapcolor( seg0, clr->idx, clr->r, clr->g, clr->b );

/* cursor stuff */
setcursor( 0, cm_truecolor(seg0,s0_2CURSOR), 0xffff );
curson();
}



/*  ************************************
    *                                  *
    *         init_globs               *
    *                                  *
    ************************************
  
    DESCRIPTION:  Initialize various globals.
  
    PREREQUSITES:  Need to run 'init_iris' first.
*/

void init_globs ( )
{
/* set up window bounds */
if ( getmonitor() == NTSC )
  {
  wminx = NTSC_XMIN;  wminy = NTSC_YMIN;
  wmaxx = NTSC_XMAX;  wmaxy = NTSC_YMAX;
  }
else if ( FALSE && mexon ) /* mex presently uses full screen */
  {
  getorigin( &wminx, &wminy );
  wmaxx += wminx - 1;
  wmaxy += wminy - 1;
  }
else
  {
  wminx = VP_XMIN;  wminy = VP_YMIN;
  wmaxx = VP_XMAX;  wmaxy = VP_YMAX;
  }

/* the screen-coord matrix */
map_rects( default_mtx, &screenrect, &ndcrect );
}



/*  ************************************
    *                                  *
    *       st_draw                    *
    *                                  *
    ************************************
  
    DESCRIPTION:  The main drawing state.
  
    ENTRY:  ignored
*/

void st_draw ( val1, val2 )
{
register pickcode;
register UNS16 thisleft;
static UNS16 prevleft=FALSE;

if ( !getbutton(ESCKEY) )
  {
  if ( (thisleft = getbutton(LEFTMOUSE))  &&  !prevleft )
    {
    if ( !check_view(v0ptr)  
        &&  !check_rot(v0ptr)
        &&  !check_xh(v0ptr) )
      readystate();
    }
  else
    readystate();

  prevleft = thisleft;
  }
}




/*  ************************************
    *                                  *
    *       readystate                 *
    *                                  *
    ************************************
  
    DESCRIPTION:
*/

void readystate ( )
{
use_view( v0ptr );
cm_color( seg1, s1_BACKG );
clear();
draw_obj();
draw_box();
draw_xh( v0ptr );
draw_view( v0ptr );
curson();
add_state( st_draw, 0, 0 );
}
