/*  FILE:  idraw.h

    DESCRIPTION:  Types, symbols, globals for 'boxview' (previously
      'idraw').
*/

#ifndef __IDRAW_H__
#define __IDRAW_H__

#include "ktypes.h"

/* recognized pipe architectures */
#define P_CLASSIC	0
#define P_GT		1
#define P_PERSONAL	2

/* default pipe architecture */
#ifndef PIPE
#  define PIPE P_CLASSIC
#endif

/* check windowing definitions */
#if PIPE==P_GT  ||  PIPE==P_PERSONAL
#  undef FOURSIGHT
#  define FOURSIGHT
#endif

/* the three directions as offsets */
#define XX 0
#define YY 1
#define ZZ 2

/* macros for accessing bits describing directions */
#define ROLLBIT(ax,dir)  ((1<<(dir))<<(ax))
#define AX_MASK(ax)      (07<<(3*(ax)))

/* types of pickable things */
#define TOK_NULL    0
#define TOKTYP_MASK 0xff00
#define TOKCOD_MASK 0x00ff
#define TOKTYP_AX   0x0100  /* axes */
#define TOKTYP_ROT  0x0200  /* rotation bars */
#define TOKTYP_VW   0x0300  /* view handles */
#define TOKTYP_XH   0x0400  /* crosshair */
#define TOKCOD_DIR  0x0003
#define TOKCOD_DIRp 0x0007
#define TOKCOD_X    XX
#define TOKCOD_Y    YY
#define TOKCOD_Z    ZZ
#define TOKCOD_p    0x0004
#define TOKCOD_Xp   (TOKCOD_p | TOKCOD_X)
#define TOKCOD_Xn   (       0 | TOKCOD_X)
#define TOKCOD_Yp   (TOKCOD_p | TOKCOD_Y)
#define TOKCOD_Yn   (       0 | TOKCOD_Y)
#define TOKCOD_Zp   (TOKCOD_p | TOKCOD_Z)
#define TOKCOD_Zn   (       0 | TOKCOD_Z)
#define TOKPT(a,b)  ( ((a&1)<<3) | ((b&1)<<2) )

/* the axes */
#define TOK_XAXIS   (TOKTYP_AX | TOKCOD_X)
#define TOK_YAXIS   (TOKTYP_AX | TOKCOD_Y)
#define TOK_ZAXIS   (TOKTYP_AX | TOKCOD_Z)

/* the pick window */
#define TOK_VEYE    (TOKTYP_VW | 0x0001)
#define TOK_VORG    (TOKTYP_VW | 0x0002)
#define TOK_VWIND   (TOKTYP_VW | 0x0003)
#define TOK_VWTOP   (TOKTYP_VW | 0x0004)
#define TOK_VWBOT   (TOKTYP_VW | 0x0005)

/* the crosshairs */
#define TOK_XH_Xp   (TOKTYP_XH | TOKCOD_Xp)
#define TOK_XH_Xn   (TOKTYP_XH | TOKCOD_Xn)
#define TOK_XH_Yp   (TOKTYP_XH | TOKCOD_Yp)
#define TOK_XH_Yn   (TOKTYP_XH | TOKCOD_Yn)
#define TOK_XH_Zp   (TOKTYP_XH | TOKCOD_Zp)
#define TOK_XH_Zn   (TOKTYP_XH | TOKCOD_Zn)
#define TOK_XH_O    (TOKTYP_XH | 0x0007)
#define TOK_XH_B    (TOKTYP_XH | 0x0008)

/* screen & pixel entities */
typedef struct SCRNPOINT  {Scoord    x,y}  SCRNPOINT;
typedef struct SCRNRECT   {SCRNPOINT  a,b}  SCRNRECT;

/* window struct */
typedef struct WINDOW
  {
  Matrix    mous_m;
  SCRNRECT  edge;
  RECT      bounds;
  RECT      bord;
  } WINDOW;

/* viewing geometry */
typedef struct VIEWGEOM
  {
  float    R;      /* distance between eye and object */
  float    D;      /* distance between eye and window */
  float    wx,wy;  /* half of height and width of window */
  } VIEWGEOM;

/* viewing data */
typedef struct VIEW
  {
  VIEWGEOM  g;
  WINDOW    *wdw;
  WINDOW    *vw;
  SCRNRECT  butt;
  XFORM     rot_to_ndc;
  XFORM     box_to_rot;
  XFORM     total;
  XFORM     vwtot;
  UNS16     viewon;  /* boolean -- TRUE if visible */
  UNS8      xhon;    /* boolean -- TRUE if crosshair visible */
  UNS8      xhdir;   /* direction of 2-D xh projection */
  VEC3      xhpos;
  } VIEW;

GLOBAL WINDOW *w0ptr;
GLOBAL VIEW *v0ptr;

GLOBAL long wminx,wmaxx,wminy,wmaxy;

/* constants */
#define VP_R    4.0
#define VP_D    2.0
#define VP_WX   1.0
#define VP_WY   0.7
#define VP_XMIN 50
#define VP_XMAX (XMAXSCREEN-50)
#define VP_YMIN 50
#define VP_YMAX (YMAXSCREEN-50)
#define NTSC_XMIN 0
#define NTSC_YMIN 0
#define NTSC_XMAX 635
#define NTSC_YMAX 484

GLOBAL UNS16 mexon;

GLOBAL Matrix default_mtx;

/* oft-used rectangles */
#ifdef CENTRAL
RECT ndcrect={ {-1,-1}, {1,1} };
RECT screenrect={ {0,0}, {XMAXSCREEN,YMAXSCREEN} };
#else
extern RECT ndcrect,screenrect;
#endif

/* font containing grab boxes */
#define NUM_GRAB_CHARS  1  /* number of characters in the grab font */
#define GRAB_FONT_IDX   1  /* font number of grab font */
#define GRAB_MAX_HEIGHT 5  /* max height of grab font character */
#define NUM_GRAB_RASTS  5  /* number of raster elements */
#ifdef CENTRAL
Fontchar grabchar[NUM_GRAB_CHARS+1] =
  {
  { 0, 0, 0,  0,  0, 0 },  /* dummy zero-th character */
  { 0, 5, 5, -1, -2, 6 }
  };
short grabrast[NUM_GRAB_RASTS] = 
  {
  /* first grab box -- 5x5 diamond */
  0x2000,
  0x7000,
  0xf800,
  0x7000,
  0x2000
  };
#else
extern Fontchar grabchar[NUM_GRAB_CHARS];
extern short grabrast[NUM_GRAB_RASTS];
#endif

/* the feedback buffer */
#if PIPE==P_PERSONAL
 typedef float FBTYPE;
#else
 typedef short FBTYPE;
#endif
#define FB_MAX 512
GLOBAL FBTYPE fb_buff[FB_MAX];

/* color indicies of various things */
#define MAX_MAPCOLOR   4095
#define CLR_BLACK      0
#define CLR_2CURSOR    1  /* 2-d cursor (little arrow) */
#define CLR_BACKG      2  /* background of everything */
#define CLR_3WDW       3  /* edge of 3-d window */
#define CLR_2WDW       4  /* edge of 2-d window */
#define CLR_AXES       5  /* axes in 3-d window */
#define CLR_AXNAMES    6  /* labels for axes */
#define CLR_VW_BACKG   8  /* background of view-manip window */
#define CLR_VW_LINES   9  /* view-manip lines */
#define CLR_VW_GRABS   10 /* view-manip handles */
#define CLR_VW_AREA    11 /* viewing area in VW window */
#define CLR_VW_BUTT    12 /* selector button for VW window */
#define CLR_SPLAT      13 /* splat color (when drawn) */
#define CLR_3CURS_EDGE 14 /* color of edge of 3-D crosshairs */
#define CLR_OBJSTART   32 /* start of object colors */

/* bitplane segment (segment 0) colors */
#define s0_CLEAR    0  /* transparent color */
#define s0_2CURSOR  1  /* 2-d cursor (little arrow) */
#define s0_3WDW     2  /* edge of 3-d window */
#define s0_2WDW     2  /* edge of 2-d window */
#define s0_AXES     3  /* axes in 3-d window */
#define s0_AXNAMES  4  /* labels for axes */
#define s0_XH_EDGE  5  /* color of edge of 3-D crosshairs */
#define s0_3FADE    6  /* faded 3-D window color */

/* other colors */
#define s1_BLACK    0  /* background black */
#define s1_BACKG    0  /* background of everything */
#define s1_VW_BACKG 0  /* background of view-manip window */
#define s1_VW_LINES 1  /* view-manip lines */
#define s1_VW_GRABS 2  /* view-manip handles */
#define s1_VW_AREA  3  /* viewing area in VW window */
#define s1_VW_BUTT  4  /* selector button for VW window */
#define s1_SPLAT    5  /* splat color (when drawn) */
#define s1_OBJSTART 10 /* start of object colors */

/* globals for two segments */
GLOBAL UNS16 seg0,seg1;

#define WROWS 20
#define WCOLS 20

#endif
