/*  FILENAME:  ktypes.h 

    DESCRIPTION:  Standard datatypes for kelvin's stuff.
*/

#ifndef KTYPES_H  /* see if we've been here before */
#define KTYPES_H  /* corresponding #endif at end of file */

/* Boolean values */
#ifndef TRUE
#define TRUE  1
#define FALSE  0
#endif

#ifdef CENTRAL
#define GLOBAL
#else
#define GLOBAL extern
#endif

/* fixed-length integer types */
typedef char INT8;            /* 8-bit signed */
typedef short INT16;          /* 16-bit signed */
typedef int INT32;            /* 32-bit signed */
typedef unsigned char UNS8;   /* 8-bit unsigned */
typedef unsigned short UNS16; /* 16-bit unsigned */
typedef unsigned int UNS32;   /* 32-bit unsigned */

#define PI      3.141592653
#define SQRT_2  1.414213562


/* say what kind of CPU this is */
#define M68000    1
#define M68020    2
#define VAX       3
#define CPU       M68000

#if CPU==M68000

  /* processor-specific integer types */
  typedef int INTREGSIZ;             /* register-sized signed */
  typedef unsigned UNSREGSIZ;        /* register-sized unsigned */
  typedef short INTDATASIZ;          /* data-bus-sized signed */
  typedef unsigned short UNSDATASIZ; /* data-bus-sized unsigned */
  typedef int INTPTRSIZ;             /* pointer-register-sized signed */
  typedef unsigned int UNSPTRSIZ;    /* pointer-register-sized unsiged */
  
  /* processor-specific sizes */
#  define DATAWIDTH  16        /* width of data bus */
#  define REGWIDTH   32        /* width of register */
#  define PTRWIDTH   24        /* meaningful width of pointer */

#endif

#if CPU==M68020

  /* processor-specific integer types */
  typedef int INTREGSIZ;           /* register-sized signed */
  typedef unsigned UNSREGSIZ;      /* register-sized unsigned */
  typedef short INTDATASIZ;        /* data-bus-sized signed */
  typedef unsigned UNSDATASIZ;     /* data-bus-sized unsigned */
  typedef int INTPTRSIZ;           /* pointer-register-sized signed */
  typedef unsigned int UNSPTRSIZ;  /* pointer-register-sized unsiged */
  
  /* processor-specific sizes */
#  define DATAWIDTH 32        /* width of data bus */
#  define REGWIDTH  32        /* width of register */
#  define PTRWIDTH  32        /* meaningful width of pointer */

#endif

#if CPU==VAX

  /* processor-specific integer types */
  typedef int INTREGSIZ;           /* register-sized signed */
  typedef unsigned UNSREGSIZ;      /* register-sized unsigned */
  typedef int INTDATASIZ;          /* data-bus-sized signed */
  typedef unsigned UNSDATASIZ;     /* data-bus-sized unsigned */
  typedef int INTPTRSIZ;           /* pointer-register-sized signed */
  typedef unsigned int UNSPTRSIZ;  /* pointer-register-sized unsiged */
  
  /* processor-specific sizes */
#  define DATAWIDTH  32        /* width of data bus */
#  define REGWIDTH   32        /* width of register */
#  define PTRWIDTH   32        /* meaningful width of pointer */

#endif

/* geometric entities */
typedef float MATRIX[4][4];
typedef struct MTX_STRUCT {MATRIX m} MTX_STRUCT;
typedef struct VEC2    {float x,y}    VEC2;
typedef struct VEC3    {float x,y,z}  VEC3;
typedef struct VEC4    {float x,y,z,w}  VEC4;
typedef struct POINT  {float x,y}    POINT;
typedef struct RECT    {POINT a,b}    RECT;
typedef union 
  {
  MATRIX     m;      /* 4x4 matrix of values */
  MTX_STRUCT a;      /* struct containing 4x4 array */
  VEC4       v[4];   /* 4 homogeneous vectors (rows) */
  float      s[16];  /* array 16 scalars */
  } XFORM;

/* constants */
#ifdef CENTRAL
  MTX_STRUCT I = 
    {{
    {1.0, 0.0, 0.0, 0.0},
    {0.0, 1.0, 0.0, 0.0},
    {0.0, 0.0, 1.0, 0.0},
    {0.0, 0.0, 0.0, 1.0}
    }};
#else
  extern MTX_STRUCT I;
#endif
#define I_INIT            \
    {                     \
    {1.0, 0.0, 0.0, 0.0}, \
    {0.0, 1.0, 0.0, 0.0}, \
    {0.0, 0.0, 1.0, 0.0}, \
    {0.0, 0.0, 0.0, 1.0}  \
    }


#endif        /* end self protection */
