/*  FILENAME: slide.c

    DESCRIPTION:  Routines to implement sliders.
*/

#include <gl.h>
#include <device.h>
#include "idraw.h"

#if PIPE==P_GT  ||  PIPE==P_PERSONAL
#  include <gl/feed.h>
#endif

/* programs in this file */
extern getslidepts();
extern void doslide();
extern void st_slide();

static SCRNRECT slidept;
static float slidex[3][3];
static void (*runfunc)(),(*donefunc)();




/*  ************************************
    *                                  *
    *        doslide()                 *
    *                                  *
    ************************************

    DESCRIPTION:

    ENTRY:
*/

void doslide ( endpt, rfunc, dfunc )
register SCRNRECT *endpt;
void (*rfunc)(),(*dfunc)();
{
register float dx,dy,denom;

/* store the endpoints and functions in the static location */
slidept = *endpt;
runfunc = rfunc;
donefunc = dfunc;

/* fix up the transform matrix */
dx = endpt->b.x - endpt->a.x;
dy = endpt->b.y - endpt->a.y;
denom = 1.0 / (dx*dx + dy*dy);
slidex[0][0] =  dx * denom;
slidex[0][1] = -dy * denom;
slidex[0][2] = 0.0;
slidex[1][0] = -slidex[0][1];
slidex[1][1] =  slidex[0][0];
slidex[1][2] = 0.0;
slidex[2][0] = -(dy * endpt->a.y + dx * endpt->a.x) * denom;
slidex[2][1] = -(dx * endpt->a.y - dy * endpt->a.x) * denom;
slidex[2][2] = 1.0;

/* perform the sliding action */
st_slide( 0, 0 );
}




/*  ************************************
    *                                  *
    *           st_slide               *
    *                                  *
    ************************************

    DESCRIPTION:

    ENTRY:
*/

void st_slide ( dum1, dum2 )
{
register float newx;
register long mx,my;

/* get mouse location */
mx = getvaluator( MOUSEX );
my = getvaluator( MOUSEY );

/* get the transformed X location */
newx = (float)mx * slidex[0][0] + (float)my * slidex[1][0] + slidex[2][0];

if ( !getbutton(LEFTMOUSE) )
  {
  (*donefunc)( newx );
  readystate();
  }
else
  {
  /* call the update routine */
  (*runfunc)( newx );
  add_state( st_slide, 0, 0 );
  }
}




/*  ************************************
    *                                  *
    *       getslidepts                *
    *                                  *
    ************************************

    DESCRIPTION:  Map a 3D line segment to a clipped screen line segment.
      Use the geometry pipeline.

    ENTRY:

    EXIT: Returns TRUE if the mapping was successful, FALSE otherwise.

    PREREQUISITES:  The viewport and transform must be set before
      calling.
*/

getslidepts ( v1, v2, srect )
register VEC3 *v1,*v2;
register SCRNRECT *srect;
{
register long fbcount;

/* check for hits */
feedback( fb_buff, FB_MAX );
  move( v1->x, v1->y, v1->z );
  draw( v2->x, v2->y, v2->z );
fbcount = endfeedback( fb_buff );

#if PIPE==P_CLASSIC
  /* make sure it's in a format we want */
  if ( fbcount != 6  ||  fb_buff[0] != 16  ||  fb_buff[3] != 17 )
    return FALSE;
  
  /* move the results to the return rectangle */
  srect->a.x = fb_buff[1];
  srect->a.y = fb_buff[2];
  srect->b.x = fb_buff[4];
  srect->b.y = fb_buff[5];
  return TRUE;
#endif

#if PIPE==P_GT
  /* make sure it's in a format we want */
  if ( fbcount != 18  ||  fb_buff[1] != 16 )
    return FALSE;
  
  /* move the results to the return rectangle */
  srect->a.x = fb_buff[2];
  srect->a.y = fb_buff[3] - 1024;
  srect->b.x = fb_buff[10];
  srect->b.y = fb_buff[11] - 1024;
  return TRUE;
#endif

#if PIPE==P_PERSONAL
  /* make sure it's in a format we want */
  if ( fbcount != 14  ||  fb_buff[0] != FB_MOVE  ||  fb_buff[7] != FB_DRAW  )
    return FALSE;
  
  /* move the results to the return rectangle */
  srect->a.x = fb_buff[1];
  srect->a.y = fb_buff[2];
  srect->b.x = fb_buff[8];
  srect->b.y = fb_buff[9];
  return TRUE;
#endif
}
