/*  FILENAME:  states.c

    DESCRIPTION:  Programs that deal with the idraw state machine.

    LINK WITH:
      tables.c -- Handles allocation of fixed-sized things.
*/

#include <gl.h>
#include <device.h>
#include "ktypes.h"

/* programs in this file */
extern void init_states();
extern void close_states();
extern do_state();

/* constants */
#define MAX_STATES  32    /* maximum number of states allowed */

/* local types */
typedef struct STATE
{
struct STATE *next;
void         (*func)();
int          val1,val2;
} STATE;

/* local variables */
static STATE
*pres_state=0,
*futr_state=0;
static
list1,list2,
futr_list;



/*  ************************************
    *                                  *
    *        init_states               *
    *                                  *
    ************************************

    DESCRIPTION:  Initialize stuff for state parser.
*/

void init_states ( )
{
list1 = init_table( MAX_STATES, sizeof(STATE) );
list2 = init_table( MAX_STATES, sizeof(STATE) );
futr_list = list1;
}




/*  ************************************
    *                                  *
    *         close_states             *
    *                                  *
    ************************************

    DESCRIPTION:  Free space used by state parser.
*/

void close_states ( )
{
close_table( list1 );
close_table( list2 );
}



/*  ************************************
    *                                  *
    *          do_state                *
    *                                  *
    ************************************

    DESCRIPTION:  Execute all state programs.
*/

do_state ( )
{
register STATE *this_state;

/* move the future to the present */
if ( futr_state == 0 )
return FALSE;
pres_state = futr_state;
futr_state = 0;

/* swap lists */
if ( futr_list == list1 )
{
reset_table( list2 );
futr_list = list2;
}
else
{
reset_table( list1 );
futr_list = list1;
}

/* loop through current states */
for ( this_state=pres_state; this_state!=0; this_state=this_state->next )
(*(this_state->func))( this_state->val1, this_state->val2 );

return TRUE;
}




/*  ************************************
    *                                  *
    *       add_state                  *
    *                                  *
    ************************************

    DESCRIPTION:  Add a state to the list to be executed in the future.

    ENTRY:
*/

void add_state ( func, val1, val2 )
void (*func)();
int val1,val2;
{
register STATE *this_state;

/* initialize */
this_state = (STATE *)get_cell( futr_list );
this_state->func = func;
this_state->val1 = val1;
this_state->val2 = val2;

/* put in linked list */
this_state->next = futr_state;
futr_state = this_state;
}
