/*  FILENAME: tables.c

    DESCRIPTION:  Utility programs for generic tables.
*/

typedef struct TABLE
  {
  unsigned short tblsize;
  unsigned short cellsize;
  char           *nextcell;
  } TABLE;

/* programs in this file */
extern int *init_table();
extern void reset_table();
extern char *get_cell();
extern void close_table();
extern void fill_table();
extern char *next_cell();




/*  *********************
    *                   *
    *    init_table     *
    *                   *
    *********************

    DESCRIPTION:  Create a table.

    ENTRY:

    EXIT:
*/

int *init_table ( tblsize, cellsize )
register tblsize;
register cellsize;
{
register TABLE *out;

/* allocate space for TABLE and init struct */
out = (TABLE *)malloc( sizeof(TABLE) + tblsize * cellsize );
out->tblsize = tblsize;
out->cellsize = cellsize;
out->nextcell = (char *)out + sizeof(TABLE) + (tblsize-1) * cellsize;

/* return pointer to struct */
return (int *)out;
}




/*  *********************
    *                   *
    *   reset_table     *
    *                   *
    *********************

    DESCRIPTION:  Initialize the elements of a table.

    ENTRY:
*/

void reset_table ( tbl )
register TABLE *tbl;
{
tbl->nextcell = (char *)(& tbl[1]) + (tbl->tblsize-1) * tbl->cellsize;
}




/*  *********************
    *                   *
    *    get_cell       *
    *                   *
    *********************

    DESCRIPTION:  Return the next cell from the table.

    ENTRY:

    EXIT:
*/

char *get_cell ( tbl )
register TABLE *tbl;
{
register char *out;

if ( tbl->nextcell != (char *)(&tbl[1]) )
  {
  out = tbl->nextcell;
  tbl->nextcell -= tbl->cellsize;
  }
else
  out = 0;

return out;
}




/*  *********************
    *                   *
    *   close_table     *
    *                   *
    *********************

    DESCRIPTION:  Destroy a table.

    ENTRY:

    EXIT:
*/

void close_table ( tbl )
register TABLE *tbl;
{
free( tbl );
}




/*  *********************
    *                   *
    *   fill_table      *
    *                   *
    *********************

    DESCRIPTION:  Initialize the elements of a table by calling
      a function for every element.

    ENTRY:
*/

void fill_table ( tbl, func )
register TABLE *tbl;
register void (*func)();
{
register idx;
register char *nextcell;

/* initialize for loop */
idx = tbl->tblsize - 1;
tbl->nextcell = nextcell
  = (char *)tbl + sizeof(TABLE) + (tbl->tblsize-1) * tbl->cellsize;

/* loop through all elements in table */
do  {
  /* call the filling function */
  (*func)( nextcell, idx );

  /* decrement counters */
  --idx;
  nextcell -= tbl->cellsize;
  }  while ( idx >= 0 );
}




/*  *********************
    *                   *
    *    next_cell      *
    *                   *
    *********************

    DESCRIPTION:  Return the next cell from the table.

    ENTRY:

    EXIT:
*/

char *next_cell ( tbl )
register TABLE *tbl;
{
register char *out;

/* get contents of next cell */
out = tbl->nextcell;

if ( tbl->nextcell == (char *)(&tbl[1]) )
  tbl->nextcell
    = (char *)tbl + sizeof(TABLE) + (tbl->tblsize-1) * tbl->cellsize;
else
  tbl->nextcell -= tbl->cellsize;

return out;
}
