/*  FILENAME: view.c

    DESCRIPTION:  Programs for manipulating viewing geometry.
*/

#include <gl.h>
#include <device.h>
#include "idraw.h"

/* programs in this file */
extern void init_view();
extern VIEW *open_view();
extern void use_view();
extern void calc_vwtot();
extern void gen_view();
extern void draw_view();
extern void draw_vmanip();
extern void draw_vbutt();
extern check_view();
extern void runwind();
extern void runwend();
extern void runeye();
extern void runorg();
extern void donevmanip();
extern void draw_main();

/* programs in 'wm.c' */
extern WINDOW *open_wdw();

/* basic constants */
#if XMAXSCREEN > 1024
#  define VW_SIZX    200   /* view-manip window X */
#  define VW_SIZY    160   /* view-manip window Y */
#  define VW_BUTTRAD 12    /* mouse cursor box size */
#else
#  define VW_SIZX    100
#  define VW_SIZY    80
#  define VW_BUTTRAD 8
#endif

/* derived constants */
#define VW_RATIO   ((float)VW_SIZY/(float)VW_SIZX)

static VIEW *now_vptr;
static VIEWGEOM prevgeom;
static float prev30,prev31,prev33;




/*  ************************************
    *                                  *
    *       init_view                  *
    *                                  *
    ************************************

    DESCRIPTION:
*/

void init_view ( )
{
}




/*  ************************************
    *                                  *
    *       open_view                  *
    *                                  *
    ************************************

    DESCRIPTION:

    EXIT:
*/

VIEW *open_view ( wdw )
register WINDOW *wdw;
{
register VIEW *vptr;

/* allocate & link */
vptr = (VIEW *)malloc( sizeof(VIEW) );
vptr->wdw = wdw;

/* viewing parameters */
vptr->g.R = VP_R;
vptr->g.D = VP_D;
vptr->g.wx = VP_WX;
vptr->g.wy = vptr->g.wx
  * (float)(wdw->edge.b.y - wdw->edge.a.y + 1)
  / (float)(wdw->edge.b.x - wdw->edge.a.x + 1);

/* matricies */
vptr->box_to_rot.a = I;
set_view( &vptr->rot_to_ndc, vptr->g.R, vptr->g.D, vptr->g.wx, vptr->g.wy );
vptr->total.a = vptr->rot_to_ndc.a;
calc_vwtot( vptr );

/* button's edges */
vptr->butt.a.x = wdw->edge.a.x;
vptr->butt.a.y = wdw->edge.a.y;
vptr->butt.b.x = vptr->butt.a.x + VW_BUTTRAD;
vptr->butt.b.y = vptr->butt.a.y + VW_BUTTRAD;

/* view-manipulation window */
vptr->vw = open_wdw( wdw->edge.a.x, wdw->edge.a.x+VW_SIZX,
  wdw->edge.a.y, wdw->edge.a.y+VW_SIZY );

vptr->viewon = FALSE;

return vptr;
}




/*  ************************************
    *                                  *
    *       use_view                   *
    *                                  *
    ************************************

    DESCRIPTION:

    ENTRY:
*/

void use_view ( vptr )
register VIEW *vptr;
{
use_wdw( vptr->wdw );
loadmatrix( & vptr->rot_to_ndc );
multmatrix( & vptr->box_to_rot );
getmatrix( & vptr->total );
}




/*  ************************************
    *                                  *
    *       reuse_view                 *
    *                                  *
    ************************************

    DESCRIPTION:

    ENTRY:
*/

void reuse_view ( vptr )
register VIEW *vptr;
{
use_wdw( vptr->wdw );
loadmatrix( & vptr->total );
}




/*  ************************************
    *                                  *
    *        check_view                *
    *                                  *
    ************************************

    DESCRIPTION:

    ENTRY:

    EXIT:
*/

check_view ( vptr )
register VIEW *vptr;
{
register long mx,my;
register pickcode;
VEC3 vec1,vec2;
SCRNRECT srect;

pickcode = 0;
mx = getvaluator( MOUSEX );
my = getvaluator( MOUSEY );

/* see if we're toggling the view window */
if ( vptr->butt.a.x <= mx  &&  mx <= vptr->butt.b.x
    &&  vptr->butt.a.y <= my  &&  my <= vptr->butt.b.y )
  {
  vptr->viewon = ! vptr->viewon;
  readystate();
  pickcode = TRUE;
  }

/* see if we're trying to change the view */
else if ( vptr->viewon )
  {
  /* check for a pick */
  vp_wdw( vptr->vw );
  loadmatrix( &vptr->vwtot );
  now_vptr = vptr;
  pickcode = check_pick( vptr->vw, &vptr->vwtot, gen_view );

  /* see if we hit any view manipulators */
  if ( (pickcode & TOKTYP_MASK) != TOKTYP_VW )
    {
    pickcode = 0;
    }
  else
    {
    /* remember the previous geometry */
    prevgeom = vptr->g;

    switch ( pickcode )
      {
      case TOK_VORG:
        /* try to map part of the X-axis to screen coords */
        vec1.x = 0.0;  vec1.y = 0.0;  vec1.z = 0.0;
        vec2.x = vptr->g.D - vptr->g.R;  vec2.y = 0.0;  vec2.z = 0.0;
        if ( !getslidepts( &vec1, &vec2, &srect ) )
          goto badpick;

        /* remember stuff */
        prev30 = vptr->vwtot.m[3][0];
        prev31 = vptr->vwtot.m[3][1];
        prev33 = vptr->vwtot.m[3][3];

        /* slide */
        doslide( &srect, runorg, donevmanip );
        break;

      case TOK_VEYE:
        /* try to map part of the X-axis to screen coords */
        vec1.x = -vptr->g.R;  vec1.y = 0.0;  vec1.z = 0.0;
        vec2.x = vptr->g.D - vptr->g.R;  vec2.y = 0.0;  vec2.z = 0.0;
        if ( !getslidepts( &vec1, &vec2, &srect ) )
          goto badpick;

        /* slide */
        doslide( &srect, runeye, donevmanip );
        break;

      case TOK_VWTOP:
        /* try to map part of the X-axis to screen coords */
        vec1.x = vptr->g.D - vptr->g.R;  vec1.y = 0.0;  vec1.z = 0.0;
        vec2.x = vec1.x;  vec2.y = vptr->g.wy;  vec2.z = 0.0;
        if ( !getslidepts( &vec1, &vec2, &srect ) )
          goto badpick;

        /* slide */
        doslide( &srect, runwend, donevmanip );
        break;

      case TOK_VWBOT:
        /* try to map part of the X-axis to screen coords */
        vec1.x = vptr->g.D - vptr->g.R;  vec1.y = 0.0;  vec1.z = 0.0;
        vec2.x = vec1.x;  vec2.y = -vptr->g.wy;  vec2.z = 0.0;
        if ( !getslidepts( &vec1, &vec2, &srect ) )
          goto badpick;

        /* slide */
        doslide( &srect, runwend, donevmanip );
        break;

      case TOK_VWIND:
        /* try to map part of the X-axis to screen coords */
        vec1.x = -vptr->g.R;  vec1.y = 0.0;  vec1.z = 0.0;
        vec2.x =        0.0;  vec2.y = 0.0;  vec2.z = 0.0;
        if ( !getslidepts( &vec1, &vec2, &srect ) )
          goto badpick;

        /* slide */
        doslide( &srect, runwind, donevmanip );
        break;

      default:
      badpick:
        ringbell();
        readystate();
        break;
      }
    }
  }

return pickcode;
}




/*  ************************************
    *                                  *
    *       draw_view                  *
    *                                  *
    ************************************

    DESCRIPTION:

    ENTRY:
*/

void draw_view ( vptr )
register VIEW *vptr;
{
if ( vptr->viewon )
  draw_vmanip( vptr );
draw_vbutt( vptr );
}





/*  ************************************
    *                                  *
    *          draw_vmanip             *
    *                                  *
    ************************************

    DESCRIPTION:  Draw view manipulation structures.

    ENTRY:
*/

void draw_vmanip ( vptr )
register VIEW *vptr;
{
register float temp;

/* set up the view window */
use_wdw( vptr->vw );
loadmatrix( & vptr->vwtot );
cm_color( seg1, s1_VW_BACKG );
clear();

/* draw the clipping box' sphere */
cm_color( seg1, s1_VW_AREA );
circf( 0.0, 0.0, SQRT_2 );

/* draw the axis */
cm_color( seg1, s1_VW_LINES );
move2( -100.0, 0.0 );
draw2(  100.0, 0.0 );

/* draw the window plane */
temp = vptr->g.D - vptr->g.R;
move2( temp, -vptr->g.wy );
draw2( temp,  vptr->g.wy );

/* draw the fostrum edges */
move2( -vptr->g.R, 0.0 );
rdr2( vptr->g.D * 100.0,  -vptr->g.wy * 100.0 );
move2( -vptr->g.R, 0.0 );
rdr2( vptr->g.D * 100.0,   vptr->g.wy * 100.0 );

/* draw the selectable markers */
font( GRAB_FONT_IDX );
cm_color( seg1, s1_VW_GRABS );
cmov2( -vptr->g.R,          0.0 );  charstr( "\001" );
cmov2(        0.0,          0.0 );  charstr( "\001" );
cmov2(        temp, -vptr->g.wy );  charstr( "\001" );
cmov2(        temp,         0.0 );  charstr( "\001" );
cmov2(        temp,  vptr->g.wy );  charstr( "\001" );
}




/*  ************************************
    *                                  *
    *        draw_vbutt                *
    *                                  *
    ************************************

    DESCRIPTION:

    ENTRY:
*/

void draw_vbutt ( vptr )
register VIEW *vptr;
{
viewport( 0, XMAXSCREEN, 0, YMAXSCREEN );
loadmatrix( default_mtx );
cm_color( seg1, s1_VW_BUTT );
rectfi( vptr->butt.a.x, vptr->butt.a.y, vptr->butt.b.x, vptr->butt.b.y );
}




/*  ************************************
    *                                  *
    *      calc_vwtot                  *
    *                                  *
    ************************************

    DESCRIPTION:

    ENTRY:
*/

#define VW_PAD  0.1

void calc_vwtot ( vptr )
register VIEW *vptr;
{
register float wantdx,wantdy,factor;
RECT ext;

/* calculate the rectangle we want to see */
ext.a.x = -(1.0+VW_PAD) * vptr->g.R;
ext.b.x =       VW_PAD  * vptr->g.R;
ext.b.y =  (1.0+VW_PAD) * vptr->g.wy;
ext.a.y = -ext.b.y;

/* calculate width, height, and ratio */
wantdx = ext.b.x - ext.a.x;
wantdy = ext.b.y - ext.a.y;

/* expand the rect to make its ratio fit */
if ( wantdy / wantdx > VW_RATIO )
  {
  /* window too tall -- need to make wider */
  factor = wantdy / (wantdx * VW_RATIO);
  ext.a.x += (1.0 - factor) * 0.5 * wantdx;
  ext.b.x = ext.a.x + wantdx * factor;
  }
else
  {
  /* window too narrow -- need to make taller */
  factor = (wantdx * VW_RATIO) / wantdy;
  ext.a.y += (1.0 - factor) * 0.5 * wantdy;
  ext.b.y = ext.a.y + wantdy * factor;
  }

map_rects( &vptr->vwtot, &ext, &ndcrect );
}




/*  ************************************
    *                                  *
    *        gen_view                  *
    *                                  *
    ************************************

    DESCRIPTION:
*/

void gen_view ( )
{
register float temp;
register VIEW *vptr;

vptr = now_vptr;
temp = vptr->g.D - vptr->g.R;

pnt2( -vptr->g.R,         0.0 );  passthrough( TOK_VEYE );
pnt2(        0.0,         0.0 );  passthrough( TOK_VORG );
pnt2(       temp, -vptr->g.wy );  passthrough( TOK_VWBOT );
pnt2(       temp,         0.0 );  passthrough( TOK_VWIND );
pnt2(       temp,  vptr->g.wy );  passthrough( TOK_VWTOP );
}




/*  ************************************
    *                                  *
    *       runwind                    *
    *                                  *
    ************************************

    DESCRIPTION:

    ENTRY:
*/

void runwind ( tt )
register float tt;
{
register VIEW *vptr;

/* initialize */
vptr = now_vptr;

/* adjust 'tt' */
if ( tt < 0.1 )
  tt = 0.1;
else if ( 0.9 < tt )
  tt = 0.9;

/* adjust the parameters */
vptr->g.D = tt * prevgeom.R;
set_view( &vptr->rot_to_ndc, vptr->g.R, vptr->g.D, vptr->g.wx, vptr->g.wy );

/* draw the main 3-box */
draw_main( vptr );
}




/*  ************************************
    *                                  *
    *         donevmanip               *
    *                                  *
    ************************************

    DESCRIPTION:

    ENTRY:
*/

void donevmanip ( tt )
register float tt;
{
calc_vwtot( now_vptr );
}




/*  ************************************
    *                                  *
    *         runwend                  *
    *                                  *
    ************************************

    DESCRIPTION:

    ENTRY:
*/

void runwend ( tt )
register float tt;
{
register VIEW *vptr;

/* initialize */
vptr = now_vptr;

/* adjust 'tt' */
if ( tt < 0.1 )
  tt = 0.1;
else if ( 2.0 < tt )
  tt = 2.0;

/* adjust the parameters */
vptr->g.wy = tt * prevgeom.wy;
vptr->g.wx = tt * prevgeom.wx;
set_view( &vptr->rot_to_ndc, vptr->g.R, vptr->g.D, vptr->g.wx, vptr->g.wy );

/* draw the main 3-box */
draw_main( vptr );
}




/*  ************************************
    *                                  *
    *           runeye                 *
    *                                  *
    ************************************

    DESCRIPTION:

    ENTRY:
*/

void runeye ( tt )
register float tt;
{
register VIEW *vptr;

/* initialize */
vptr = now_vptr;

/* adjust 'tt' */
if ( tt < -2.0 )
  tt = -2.0;
else if ( tt > 0.9 )
  tt = 0.9;

/* adjust the parameters */
vptr->g.R = prevgeom.R - tt * prevgeom.D;
vptr->g.D = prevgeom.D - (prevgeom.R - vptr->g.R);
set_view( &vptr->rot_to_ndc, vptr->g.R, vptr->g.D, vptr->g.wx, vptr->g.wy );

/* draw the main 3-box */
draw_main( vptr );
}




/*  ************************************
    *                                  *
    *           runorg                 *
    *                                  *
    ************************************

    DESCRIPTION:

    ENTRY:
*/

void runorg ( tt )
register float tt;
{
register VIEW *vptr;
register float dx;

/* initialize */
vptr = now_vptr;

/* adjust 'tt' */
if ( tt < -2.0 )
  tt = -2.0;
else if ( tt > 0.9 )
  tt = 0.9;

/* adjust the viewing geometry */
vptr->g.R = prevgeom.R - tt * (prevgeom.R - prevgeom.D);
set_view( &vptr->rot_to_ndc, vptr->g.R, vptr->g.D, vptr->g.wx, vptr->g.wy );

/* adjust the transform for the view window */
dx = prevgeom.R - vptr->g.R;
vptr->vwtot.m[3][0] = prev30 - dx * vptr->vwtot.m[0][0];
vptr->vwtot.m[3][1] = prev31 - dx * vptr->vwtot.m[0][1];
vptr->vwtot.m[3][3] = prev33 - dx * vptr->vwtot.m[0][3];

/* draw the main 3-box */
draw_main( vptr );
}




/*  ************************************
    *                                  *
    *         draw_main                *
    *                                  *
    ************************************

    DESCRIPTION:

    ENTRY:
*/

void draw_main ( vptr )
VIEW *vptr;
{
use_view( vptr );
cm_color( seg1, s1_BACKG );
clear();
draw_obj();
draw_box();
draw_vmanip( vptr );
}
