/*  FILENAME: wm.c

    DESCRIPTION:  Window manager routines.
*/

#include <gl.h>
#include <device.h>
#include "idraw.h"


/* programs in this file */
extern char *open_wdw();
extern void use_wdw();
extern void vp_wdw();
extern void close_wdw();
extern getmousebox();

#define PICK_RAD 4

static WINDOW *curr_wdw;




/*  ************************************
    *                                  *
    *      open_wdw                    *
    *                                  *
    ************************************

    DESCRIPTION:  Open a window on the screen.

    ENTRY:

    EXIT:
*/

char *open_wdw ( minx, maxx, miny, maxy )
long minx,maxx,miny,maxy;
{
register WINDOW *wdw;

wdw = (WINDOW *)malloc( sizeof(WINDOW) );

if ( mexon )
  {
  getorigin( &wminx, &wminy );
  getsize( &wmaxx, &wmaxy );
  wmaxx += wminx - 1;
  wmaxy += wminy - 1;
  minx = wminx;  miny = wminy;
  maxx = wmaxx;  maxy = wmaxy;
  }

/* record bounds of window */
wdw->edge.a.x = minx;
wdw->edge.a.y = miny;
wdw->edge.b.x = maxx;
wdw->edge.b.y = maxy;
wdw->bounds.a.x = minx;
wdw->bounds.a.y = miny;
wdw->bounds.b.x = maxx;
wdw->bounds.b.y = maxy;
wdw->bord.a.x = minx-1;
wdw->bord.a.y = miny-1;
wdw->bord.b.x = maxx+1;
wdw->bord.b.y = maxy+1;

/* calculate mapping from screen to window */
map_rects( wdw->mous_m, &wdw->bounds, &ndcrect );

return (char *)wdw;
}




/*  ************************************
    *                                  *
    *       close_wdw                  *
    *                                  *
    ************************************

    DESCRIPTION:

    ENTRY:
*/

void close_wdw ( wdw )
{
free( wdw );
}



/*  ************************************
    *                                  *
    *         use_wdw                  *
    *                                  *
    ************************************

    DESCRIPTION:  Set the screen mask and current transform to
      reflect a given window.

    ENTRY:
*/

void use_wdw ( wdw )
register WINDOW *wdw;
{
/* draw the window's border */
viewport( 0, XMAXSCREEN, 0, YMAXSCREEN );
loadmatrix( default_mtx );
cm_color( seg0, s0_2WDW );
rect( wdw->bord.a.x, wdw->bord.a.y, wdw->bord.b.x, wdw->bord.b.y );

viewport( wdw->edge.a.x, wdw->edge.b.x, wdw->edge.a.y, wdw->edge.b.y );

curr_wdw = wdw;
}



/*  ************************************
    *                                  *
    *          vp_wdw                  *
    *                                  *
    ************************************

    DESCRIPTION:  Set the viewport to the indicated window.

    ENTRY:
*/

void vp_wdw ( wdw )
register WINDOW *wdw;
{
viewport( wdw->edge.a.x, wdw->edge.b.x, wdw->edge.a.y, wdw->edge.b.y );
curr_wdw = wdw;
}




/*  ************************************
    *                                  *
    *      clear_wdw                   *
    *                                  *
    ************************************

    DESCRIPTION:

    ENTRY:
*/

void clear_wdw ( wdw )
register WINDOW *wdw;
{
pushmatrix();
  loadmatrix( default_mtx );
  rectfi( wdw->edge.a.x, wdw->edge.a.y, wdw->edge.b.x, wdw->edge.b.y );
popmatrix();
}




/*  ************************************
    *                                  *
    *      getmousebox                 *
    *                                  *
    ************************************

    DESCRIPTION:

    ENTRY:

    EXIT:
*/

getmousebox ( wdw, boxptr )
register WINDOW *wdw;
register RECT *boxptr;
{
register long mx,my;

/* init */
mx = getvaluator(MOUSEX);
my = getvaluator(MOUSEY);

/* make sure it's inside the window */
if ( mx < wdw->edge.a.x  ||  wdw->edge.b.x < mx
    ||  my < wdw->edge.a.y  ||  wdw->edge.b.y < my )
  return FALSE;

/* do the transform to window coords by hand */
boxptr->a.x = (mx-PICK_RAD) * wdw->mous_m[0][0] + wdw->mous_m[3][0];
boxptr->a.y = (my-PICK_RAD) * wdw->mous_m[1][1] + wdw->mous_m[3][1];
boxptr->b.x = (mx+PICK_RAD) * wdw->mous_m[0][0] + wdw->mous_m[3][0];
boxptr->b.y = (my+PICK_RAD) * wdw->mous_m[1][1] + wdw->mous_m[3][1];

return TRUE;
}
