/*  FILENAME: xform.c

    DESCRIPTION:  Various matrix processors for idraw.
*/

#include <math.h>
#include <gl.h>
#include "ktypes.h"
#include "idraw.h"

/* programs in this file */
extern XFORM *mmul();
extern VEC4 *vmul();
/* extern void set_vp(); */
extern void rot_x();
extern void rot_y();
extern void rot_z();
extern void set_view();
extern void map_rects(),qmap_rects();
extern unsigned getdirs();
extern XFORM *axslopes();

#define MINISCULE 1.0e-32




/*  *********************
    *                   *
    *      mmul         *
    *                   *
    *********************

    DESCRIPTION:  Multiply two 4x4 matricies.

    ENTRY:
      a,b -- two matricies to be multiplied
      result -- result matrix, perhaps the same as an input matrix
*/

XFORM *mmul ( result, a, b )
register XFORM *a,*b,*result;
{
XFORM tempx;

tempx.m[0][0] =	a->m[0][0] * b->m[0][0]
	     +  a->m[0][1] * b->m[1][0]
	     +  a->m[0][2] * b->m[2][0]
	     +  a->m[0][3] * b->m[3][0];

tempx.m[0][1] =	a->m[0][0] * b->m[0][1]
	     +  a->m[0][1] * b->m[1][1]
	     +  a->m[0][2] * b->m[2][1]
	     +  a->m[0][3] * b->m[3][1];

tempx.m[0][2] =	a->m[0][0] * b->m[0][2]
	     +  a->m[0][1] * b->m[1][2]
	     +  a->m[0][2] * b->m[2][2]
	     +  a->m[0][3] * b->m[3][2];

tempx.m[0][3] =	a->m[0][0] * b->m[0][3]
	     +  a->m[0][1] * b->m[1][3]
	     +  a->m[0][2] * b->m[2][3]
	     +  a->m[0][3] * b->m[3][3];

tempx.m[1][0] =	a->m[1][0] * b->m[0][0]
	     +  a->m[1][1] * b->m[1][0]
	     +  a->m[1][2] * b->m[2][0]
	     +  a->m[1][3] * b->m[3][0];

tempx.m[1][1] =	a->m[1][0] * b->m[0][1]
	     +  a->m[1][1] * b->m[1][1]
	     +  a->m[1][2] * b->m[2][1]
	     +  a->m[1][3] * b->m[3][1];

tempx.m[1][2] =	a->m[1][0] * b->m[0][2]
	     +  a->m[1][1] * b->m[1][2]
	     +  a->m[1][2] * b->m[2][2]
	     +  a->m[1][3] * b->m[3][2];

tempx.m[1][3] =	a->m[1][0] * b->m[0][3]
	     +  a->m[1][1] * b->m[1][3]
	     +  a->m[1][2] * b->m[2][3]
	     +  a->m[1][3] * b->m[3][3];

tempx.m[2][0] =	a->m[2][0] * b->m[0][0]
	     +  a->m[2][1] * b->m[1][0]
	     +  a->m[2][2] * b->m[2][0]
	     +  a->m[2][3] * b->m[3][0];

tempx.m[2][1] =	a->m[2][0] * b->m[0][1]
	     +  a->m[2][1] * b->m[1][1]
	     +  a->m[2][2] * b->m[2][1]
	     +  a->m[2][3] * b->m[3][1];

tempx.m[2][2] =	a->m[2][0] * b->m[0][2]
	     +  a->m[2][1] * b->m[1][2]
	     +  a->m[2][2] * b->m[2][2]
	     +  a->m[2][3] * b->m[3][2];

tempx.m[2][3] =	a->m[2][0] * b->m[0][3]
	     +  a->m[2][1] * b->m[1][3]
	     +  a->m[2][2] * b->m[2][3]
	     +  a->m[2][3] * b->m[3][3];

tempx.m[3][0] =	a->m[3][0] * b->m[0][0]
	     +  a->m[3][1] * b->m[1][0]
	     +  a->m[3][2] * b->m[2][0]
	     +  a->m[3][3] * b->m[3][0];

tempx.m[3][1] =	a->m[3][0] * b->m[0][1]
	     +  a->m[3][1] * b->m[1][1]
	     +  a->m[3][2] * b->m[2][1]
	     +  a->m[3][3] * b->m[3][1];

tempx.m[3][2] =	a->m[3][0] * b->m[0][2]
	     +  a->m[3][1] * b->m[1][2]
	     +  a->m[3][2] * b->m[2][2]
	     +  a->m[3][3] * b->m[3][2];

tempx.m[3][3] =	a->m[3][0] * b->m[0][3]
	     +  a->m[3][1] * b->m[1][3]
	     +  a->m[3][2] * b->m[2][3]
	     +  a->m[3][3] * b->m[3][3];

result->a = tempx.a;	/* copy temporary product to result matrix */
return result;
}



/*  ************************************
    *                                  *
    *         vmul                     *
    *                                  *
    ************************************

    DESCRIPTION:  Multiply a row vector by a matrix.

    ENTRY:

    EXIT:
*/

VEC4 *vmul ( result, v, m )
register XFORM *m;
register VEC4 *result,*v;
{
VEC4 tempv;

tempv.x =   v->x * m->m[0][0]
	 +  v->y * m->m[1][0]
	 +  v->z * m->m[2][0]
	 +  v->w * m->m[3][0];

tempv.y =   v->x * m->m[0][1]
	 +  v->y * m->m[1][1]
	 +  v->z * m->m[2][1]
	 +  v->w * m->m[3][1];

tempv.z =   v->x * m->m[0][2]
	 +  v->y * m->m[1][2]
	 +  v->z * m->m[2][2]
	 +  v->w * m->m[3][2];

tempv.w =   v->x * m->m[0][3]
	 +  v->y * m->m[1][3]
	 +  v->z * m->m[2][3]
	 +  v->w * m->m[3][3];

*result = tempv;
return result;
}



/*  ************************************
    *                                  *
    *         rot_x                    *
    *                                  *
    ************************************

    DESCRIPTION:  Add an X rotation to a matrix.

    ENTRY:
*/

void rot_x ( matx, ang )
register XFORM *matx;
float ang;
{
XFORM rotmatx;
register float sinx,cosx;

/* initialize */
sinx = sin(ang);
cosx = cos(ang);

/* set up the local matrix */
rotmatx.a = I;
rotmatx.m[1][1] =  cosx;  rotmatx.m[1][2] =  sinx;
rotmatx.m[2][1] = -sinx;  rotmatx.m[2][2] =  cosx;

/* multiply */
mmul( matx, &rotmatx, matx );
}



/*  ************************************
    *                                  *
    *         rot_y                    *
    *                                  *
    ************************************

    DESCRIPTION:  Add a Y rotation to a matrix.

    ENTRY:
*/

void rot_y ( matx, ang )
register XFORM *matx;
float ang;
{
XFORM rotmatx;
register float sinx,cosx;

/* initialize */
sinx = sin(ang);
cosx = cos(ang);

/* set up the local matrix */
rotmatx.a = I;
rotmatx.m[0][0] =  cosx;  rotmatx.m[0][2] = -sinx;
rotmatx.m[2][0] =  sinx;  rotmatx.m[2][2] =  cosx;

/* multiply */
mmul( matx, &rotmatx, matx );
}



/*  ************************************
    *                                  *
    *         rot_z                    *
    *                                  *
    ************************************

    DESCRIPTION:  Add a Z rotation to a matrix.

    ENTRY:
*/

void rot_z ( matx, ang )
register XFORM *matx;
float ang;
{
XFORM rotmatx;
register float sinx,cosx;

/* initialize */
sinx = sin(ang);
cosx = cos(ang);

/* set up the local matrix */
rotmatx.a = I;
rotmatx.m[0][0] =  cosx;  rotmatx.m[0][1] =  sinx;
rotmatx.m[1][0] = -sinx;  rotmatx.m[1][1] =  cosx;

/* multiply */
mmul( matx, &rotmatx, matx );
}




/*  ************************************
    *                                  *
    *        set_view                  *
    *                                  *
    ************************************

    DESCRIPTION:

    ENTRY:
*/

void set_view ( matx, r, d, wx, wy )
register XFORM *matx;
register float r,d,wx,wy;
{
matx->a = I;
matx->m[0][0] = d / wx;
matx->m[1][1] = d / wy;
matx->m[2][2] = 0.0;
matx->m[2][3] = -1.0;
matx->m[3][3] = r;
}




/*  ************************************
    *                                  *
    *        map_rects                 *
    *                                  *
    ************************************

    DESCRIPTION:

    ENTRY:
*/

void map_rects ( m, from, to )
register MTX_STRUCT *m;
register RECT *from,*to;
{
register float fromdiff,todiff,ratio;

*m = I;

/* figure X transform */
fromdiff = from->b.x - from->a.x;
todiff = to->b.x - to->a.x;
ratio = todiff / fromdiff;
m->m[0][0] = ratio;
m->m[3][0] = to->a.x - from->a.x * ratio;

/* figure Y transform */
fromdiff = from->b.y - from->a.y;
todiff = to->b.y - to->a.y;
ratio = todiff / fromdiff;
m->m[1][1] = ratio;
m->m[3][1] = to->a.y - from->a.y * ratio;
}




/*  ************************************
    *                                  *
    *        qmap_rects                *
    *                                  *
    ************************************

    DESCRIPTION:

    ENTRY:
*/

void qmap_rects ( m, from, to )
register MTX_STRUCT *m;
register RECT *from,*to;
{
register float fromdiff,todiff,ratio;

/* figure X transform */
fromdiff = from->b.x - from->a.x;
todiff = to->b.x - to->a.x;
ratio = todiff / fromdiff;
m->m[0][0] = ratio;
m->m[3][0] = to->a.x - from->a.x * ratio;

/* figure Y transform */
fromdiff = from->b.y - from->a.y;
todiff = to->b.y - to->a.y;
ratio = todiff / fromdiff;
m->m[1][1] = ratio;
m->m[3][1] = to->a.y - from->a.y * ratio;
}



/*  ************************************
    *                                  *
    *        getdirs                   *
    *                                  *
    ************************************

    DESCRIPTION:  Find out which way coordinate axes are
      pointing on the screen.

    ENTRY:
      mtx -- pointer to an XFORM struct describing
        the world-to-eye transform

    EXIT: Returns 9 bits in an 'unsigned'.  The bits may be
      read via the ROLLBIT macro (in 'idraw.h').

    DERIVATION:
        We want to find out if the positive end of axis A maps
      to a positive D component in eye space.
        Assume we have a general 4x4 transform matrix 'mtx'.
      To see which way the X axis is pointing, we send the
      vector  [+inf 0 0 1]  throuh 'mtx' and see what comes out.
      The signs of things in the result vector tell us which way
      the axes are pointing.  Because of the infinite component
      in the input vector, the bottom row of 'mtx' doesn't
      matter.  Each component of the result vector is a ratio of
      two infinities, but we only care about the sign of each ratio.
      The sign of a ratio is the XOR of the signs of its numerator
      and denominator.
*/


unsigned getdirs ( mtx )
register XFORM *mtx;
{
register unsigned num,den;

/* init */
num = 0;  den = 0;

#define DO_NUM(ax,dir) \
  {if ( mtx->m[ax][dir] > 0.0 )  num |= ROLLBIT(ax,dir);}
DO_NUM(XX,XX);  DO_NUM(XX,YY);  DO_NUM(XX,ZZ);
DO_NUM(YY,XX);  DO_NUM(YY,YY);  DO_NUM(YY,ZZ);
DO_NUM(ZZ,XX);  DO_NUM(ZZ,YY);  DO_NUM(ZZ,ZZ);

#define DO_DEN(ax) \
  {if ( mtx->m[ax][3] > 0.0 )  den |= AX_MASK(ax);}
DO_DEN(XX);
DO_DEN(YY);
DO_DEN(ZZ);

return num ^ den;
}




/*  ************************************
    *                                  *
    *       axslopes                   *
    *                                  *
    ************************************

    DESCRIPTION:  Find out the slopes of the coordinate axes
      in eye space.

    ENTRY:
      mtx -- Pointer to an XFORM union containing the current
        world-to-eye transform.
      slope -- Pointer to another XFORM union to receive the
        slope matrix.  Only the upper left 3x3 submatrix
        will be filled, where each row describes the slope
        of an axis.

    EXIT:  Returns the 'slope' pointer.
*/

XFORM *axslopes ( mtx, slope )
register XFORM *mtx,*slope;
{
register float denom;

#define XXPPQQ(ax,dir)                    \
  slope->m[ax][dir] =                     \
    (mtx->m[ax][dir] + mtx->m[3][dir])    \
      / (mtx->m[ax][3] + mtx->m[3][3]);
#define DO_AX_SLP(ax) \
  { XXPPQQ(ax,XX);  XXPPQQ(ax,YY);  XXPPQQ(ax,ZZ); }

DO_AX_SLP(XX);
DO_AX_SLP(YY);
DO_AX_SLP(ZZ);

return slope;
}
