/*  FILENAME: xh.c

    DESCRIPTION:  Deal with the crosshairs
*/


#include <stdio.h>
#include <math.h>
#include <gl.h>
#include <device.h>
#include "idraw.h"

/* programs in this file */
extern void init_xh();
extern void open_xh();
extern void draw_xh();
extern check_xh();
extern void st_xhmov2();
extern void map2d3d();
/*extern unsigned mostest(); */

/* programs in other files */
extern void st_axis();

#define EDGE_FLAG  (TOKCOD_p << 1)
#define MOUSE_MAX  512
#define FABS(x)    fabs(x)

typedef struct MOUSEFLAT
  {
  VIEW  *vptr;
  int    xsign,ysign;
  } MOUSEFLAT;




/*  ************************************
    *                                  *
    *        init_xh                   *
    *                                  *
    ************************************

    DESCRIPTION:  Initialize crosshairs.
*/

void init_xh ( )
{
}




/*  ************************************
    *                                  *
    *        open_xh                   *
    *                                  *
    ************************************

    DESCRIPTION:  Open crosshairs.

    ENTRY:
      vptr -- Pointer to VIEW struct to hold crosshair info.
*/

void open_xh ( vptr )
register VIEW *vptr;
{
vptr->xhon = TRUE;
vptr->xhdir = FALSE;
vptr->xhpos.x = 0.0;
vptr->xhpos.y = 0.0;
vptr->xhpos.z = 0.0;
}




/*  ************************************
    *                                  *
    *         draw_xh                  *
    *                                  *
    ************************************

    DESCRIPTION:  Draw the 3-D crosshairs.

    ENTRY:

    EXIT:
*/

void draw_xh ( vptr )
register VIEW *vptr;
{
if ( vptr->xhon )
  {
  /* crosshair color */
  cm_color( seg0, s0_AXES );

  /* X hair */
  move(          -1.0, vptr->xhpos.y, vptr->xhpos.z );
  draw(           1.0, vptr->xhpos.y, vptr->xhpos.z );

  /* Y hair */
  move( vptr->xhpos.x,          -1.0, vptr->xhpos.z );
  draw( vptr->xhpos.x,           1.0, vptr->xhpos.z );

  /* Z hair */
  move( vptr->xhpos.x, vptr->xhpos.y,          -1.0 );
  draw( vptr->xhpos.x, vptr->xhpos.y,           1.0 );

  /* edges */
  if ( vptr->xhdir )
    {
    cm_color( seg0, s0_XH_EDGE );
    switch ( vptr->xhdir & TOKCOD_DIR )
      {
      case TOKCOD_X:
        move( vptr->xhpos.x, -1.0, -1.0 );
        draw( vptr->xhpos.x,  1.0, -1.0 );
        draw( vptr->xhpos.x,  1.0,  1.0 );
        draw( vptr->xhpos.x, -1.0,  1.0 );
        draw( vptr->xhpos.x, -1.0, -1.0 );
        break;

      case TOKCOD_Y:
        move( -1.0, vptr->xhpos.y, -1.0 );
        draw( -1.0, vptr->xhpos.y,  1.0 );
        draw(  1.0, vptr->xhpos.y,  1.0 );
        draw(  1.0, vptr->xhpos.y, -1.0 );
        draw( -1.0, vptr->xhpos.y, -1.0 );
        break;

      case TOKCOD_Z:
        move( -1.0, -1.0, vptr->xhpos.z );
        draw(  1.0, -1.0, vptr->xhpos.z );
        draw(  1.0,  1.0, vptr->xhpos.z );
        draw( -1.0,  1.0, vptr->xhpos.z );
        draw( -1.0, -1.0, vptr->xhpos.z );
        break;
      }
    }

  /* labels */
  font( 0 );
  cm_color( seg0, s0_AXNAMES );
  cmov(            1.0, vptr->xhpos.y, vptr->xhpos.z );  charstr( "X" );
  cmov(  vptr->xhpos.x,           1.0, vptr->xhpos.z );  charstr( "Y" );
  cmov(  vptr->xhpos.x, vptr->xhpos.y,           1.0 );  charstr( "Z" );
  }
}




/*  ************************************
    *                                  *
    *           gen_xh                 *
    *                                  *
    ************************************

    DESCRIPTION:  Generate feedback instructions for
      3D crosshairs.

    ENTRY:
*/

void gen_xh ( vptr )
register VIEW *vptr;
{
move(           0.0, vptr->xhpos.y, vptr->xhpos.z );
draw(          -1.0, vptr->xhpos.y, vptr->xhpos.z );
passthrough( TOK_XH_Xn );

move(           0.0, vptr->xhpos.y, vptr->xhpos.z );
draw(           1.0, vptr->xhpos.y, vptr->xhpos.z );
passthrough( TOK_XH_Xp );

move( vptr->xhpos.x,           0.0, vptr->xhpos.z );
draw( vptr->xhpos.x,          -1.0, vptr->xhpos.z );
passthrough( TOK_XH_Yn );

move( vptr->xhpos.x,           0.0, vptr->xhpos.z );
draw( vptr->xhpos.x,           1.0, vptr->xhpos.z );
passthrough( TOK_XH_Yp );

move( vptr->xhpos.x, vptr->xhpos.y,           0.0 );
draw( vptr->xhpos.x, vptr->xhpos.y,          -1.0 );
passthrough( TOK_XH_Zn );

move( vptr->xhpos.x, vptr->xhpos.y,           0.0 );
draw( vptr->xhpos.x, vptr->xhpos.y,           1.0 );
passthrough( TOK_XH_Zp );
}




/*  ************************************
    *                                  *
    *       check_xh                   *
    *                                  *
    ************************************

    DESCRIPTION:  See if mouse over crosshairs.

    ENTRY:

    EXIT:  Returns status code.
*/

check_xh ( vptr )
register VIEW *vptr;
{
register UNS16 pickcode;
UNS16 pos;
register WINDOW *wdw;
register out;
short xinit,yinit;
static MOUSEFLAT thisflat;

wdw = vptr->wdw;
viewport( wdw->edge.a.x, wdw->edge.b.x, wdw->edge.a.y, wdw->edge.b.y );
pickcode = check_pick( vptr->wdw, &vptr->total, gen_xh, vptr );

if ( (pickcode & TOKTYP_MASK) != TOKTYP_XH )
  {
  out = FALSE;
  }
else
  {
  /* init */
  pos = pickcode & TOKCOD_p;
  vptr->xhdir = EDGE_FLAG | (pickcode & TOKCOD_DIRp);

  /* redraw screen */
  cursoff();
  draw_rot( vptr );

  /* figure mapping of mouse to 3d crosshairs */
  thisflat.vptr = vptr;
  thisflat.xsign = 1;
  thisflat.ysign = 1;
  map2d3d( & thisflat );  /* modifies 'thisflat' */

  /* reset the mouse scaling */
  switch ( vptr->xhdir & TOKCOD_DIRp )
    {
    case TOKCOD_Xp:
      xinit = vptr->xhpos.y * MOUSE_MAX;
      yinit = vptr->xhpos.z * MOUSE_MAX;
      break;
    case TOKCOD_Xn:
      xinit = vptr->xhpos.z * MOUSE_MAX;
      yinit = vptr->xhpos.y * MOUSE_MAX;
      break;
    case TOKCOD_Yp:
      xinit = vptr->xhpos.z * MOUSE_MAX;
      yinit = vptr->xhpos.x * MOUSE_MAX;
      break;
    case TOKCOD_Yn:
      xinit = vptr->xhpos.x * MOUSE_MAX;
      yinit = vptr->xhpos.z * MOUSE_MAX;
      break;
    case TOKCOD_Zp:
      xinit = vptr->xhpos.x * MOUSE_MAX;
      yinit = vptr->xhpos.y * MOUSE_MAX;
      break;
    case TOKCOD_Zn:
      xinit = vptr->xhpos.y * MOUSE_MAX;
      yinit = vptr->xhpos.x * MOUSE_MAX;
      break;
    }
  xinit *= thisflat.xsign;
  yinit *= thisflat.ysign;
  setvaluator( MOUSEX, xinit, -MOUSE_MAX,  MOUSE_MAX );
  setvaluator( MOUSEY, yinit, -MOUSE_MAX,  MOUSE_MAX );

  /* set new state */
  add_state( st_xhmov2, &thisflat, 0 );

  out = TRUE;
  }

return out;
}




/*  ************************************
    *                                  *
    *          st_xhmov2               *
    *                                  *
    ************************************

    DESCRIPTION:  Move the crosshair position in 2D.

    ENTRY:
      axis -- code for frozen axis
*/

void st_xhmov2 ( thisflat, dum )
register MOUSEFLAT *thisflat;
{
register VIEW *vptr;
register float xpos,ypos;

/* init */
vptr = thisflat->vptr;

if ( ! getbutton(LEFTMOUSE) )
  {
  VEC4 xhwld,xhscn;
  XFORM toscreen;
  short xx,yy;

  /* reset variables */
  vptr->xhdir = FALSE;

  /* */
  map_rects( &toscreen, &ndcrect, &vptr->wdw->bounds );
  mmul( &toscreen, &vptr->total, &toscreen );
  xhwld.x = vptr->xhpos.x;
  xhwld.y = vptr->xhpos.y;
  xhwld.z = vptr->xhpos.z;
  xhwld.w = 1.0;
  vmul( &xhscn, &xhwld, &toscreen );
  xx = xhscn.x / xhscn.w;
  yy = xhscn.y / xhscn.w;

  /* reset mouse scaling */
  setvaluator( MOUSEX, xx, 0, XMAXSCREEN );
  setvaluator( MOUSEY, yy, 0, YMAXSCREEN );
  curson();

  /* go to base state */
  readystate();
  }
else
  {
  /* get 2D real mouse position */
  cursoff();
  xpos = thisflat->xsign * getvaluator( MOUSEX ) * (1.0/(float)MOUSE_MAX);
  ypos = thisflat->ysign * getvaluator( MOUSEY ) * (1.0/(float)MOUSE_MAX);

  /* change the position */
  switch ( vptr->xhdir & TOKCOD_DIRp )
    {
    case TOKCOD_Xp:
      vptr->xhpos.z = ypos;
      vptr->xhpos.y = xpos;
      break;

    case TOKCOD_Xn:
      vptr->xhpos.z = xpos;
      vptr->xhpos.y = ypos;
      break;

    case TOKCOD_Yp:
      vptr->xhpos.z = xpos;
      vptr->xhpos.x = ypos;
      break;

    case TOKCOD_Yn:
      vptr->xhpos.z = ypos;
      vptr->xhpos.x = xpos;
      break;

    case TOKCOD_Zp:
      vptr->xhpos.x = xpos;
      vptr->xhpos.y = ypos;
      break;

    case TOKCOD_Zn:
      vptr->xhpos.x = ypos;
      vptr->xhpos.y = xpos;
      break;
    }

  draw_rot( v0ptr );
  add_state( st_xhmov2, thisflat, 0 );
  }
}



/*  ************************************
    *                                  *
    *        map2d3d                   *
    *                                  *
    ************************************

    DESCRIPTION:  Fill 'vptr->xhdir' and '?sign' in
      a MOUSEFLAT struct to determine how the mouse moves
      the 3D crosshairs.

    ENTRY:
*/

#define SLPD(ax,dir) FABS(axslp[ax][dir])
#define MOSTER(ax1,ax2,dir)                                         \
  (                                                                 \
    ( SLPD(ax1,dir) > SLPD(ax2,dir) )                               \
      ? ( (axslp[ax1][dir] > 0.0)  ?  (ax1) | TOKCOD_p  :  (ax1) )  \
      : ( (axslp[ax2][dir] > 0.0)  ?  (ax2) | TOKCOD_p  :  (ax2) )  \
  )

void map2d3d ( fptr )
register MOUSEFLAT *fptr;
{
register VIEW *vptr;
register XFORM *total;
UNS16 upaxis,rtaxis;
Matrix axslp;  /* axis slopes */

/* init */
vptr = fptr->vptr;
total = & vptr->total;
vptr->xhdir &= ~TOKCOD_p;

/* discover arrangement of axes on screen */
axslopes( total, axslp );

switch ( vptr->xhdir & TOKCOD_DIR )
  {
  case TOKCOD_Z:
    rtaxis = MOSTER( XX, YY, XX );
    upaxis = MOSTER( XX, YY, YY );
    if ( (rtaxis & TOKCOD_DIR) == XX )
      vptr->xhdir |= TOKCOD_p;
    break;

  case TOKCOD_X:
    rtaxis = MOSTER( YY, ZZ, XX );
    upaxis = MOSTER( YY, ZZ, YY );
    if ( (rtaxis & TOKCOD_DIR) == YY )
      vptr->xhdir |= TOKCOD_p;
    break;

  case TOKCOD_Y:
    rtaxis = MOSTER( XX, ZZ, XX );
    upaxis = MOSTER( XX, ZZ, YY );
    if ( (rtaxis & TOKCOD_DIR) == ZZ )
      vptr->xhdir |= TOKCOD_p;
    break;
  }

fptr->xsign = (rtaxis & TOKCOD_p) ?  1 : -1;
fptr->ysign = (upaxis & TOKCOD_p) ?  1 : -1;
}
